@@ -20,11 +20,17 @@
 #ifndef COUNTEREXAMPLE_H
 # define COUNTEREXAMPLE_H
 
-# include "state-item.h"
+# include "state.h"
 
+// Init/deinit this module.
 void counterexample_init (void);
 void counterexample_free (void);
 
-void counterexample_report_state (const state *s, FILE *out, const char *prefix);
+// Print the counterexamples for the conflicts of state S.
+//
+// Used both for the warnings on the terminal (OUT = stderr, PREFIX =
+// ""), and for the reports (OUT != stderr, PREFIX != "").
+void
+counterexample_report_state (const state *s, FILE *out, const char *prefix);
 
 #endif /* COUNTEREXAMPLE_H */
